

// Paul Johnson <pauljohn@ku.edu> 2002
// This code is made available under the GPL (GNU general public license)
// with the additional proviso that any users notify the author of their
// use and share their code with the original author.  A copy of version
// 2 of the GPL is included in this directory, but you can use a newer
// GPL at your choice.  The author assumes no liability for any 
// inconvenience or damage caused by this software and offers no 
// warranty of any kind.


//Adapted from "PermutedIndex" by Marcus Daniels, swarm-support, June 21, 2000
//avltr is "right threaded" avl tree. Better if you need "next" support.

#import <simtools.h> // initSwarm
#import <collections.h> // Array
#import <defobj/Create.h> // CreateDrop
#import <objectbase/Swarm.h> // Swarm
#import <defobj/defalloc.h> // getZone
#import <objectbase/SwarmObject.h>  //superclass of AVLtree

#include <collections/Collection.h> 

#include <misc/avl.h> //for the AVLtree
     
typedef int (*compare_avl_t) (const void *, const void *,void *);


@interface AVLSet: SwarmObject <KeyedCollection>
{
  avl_tree * aTree;
  compare_avl_t  compareFunc;
  BOOL replaceOnly;
}

+ createBegin: aZone;

+ createBegin:  (int (*)(const void*,const void*,void*)) compareFunc  Param: (void *) param Zone: aZone;

- setCompareFunction: (compare_avl_t) compareFunction;

- setCompareCStrings;

- setCompareIDs;

- setCompareIntegers;

- setCompareUnsignedIntegers;

- setReplaceOnly: (BOOL) condition;

- (BOOL) getReplaceOnly;

- createEnd;

- copy: aZone;

- (int) count;

- (unsigned) getCount;

-(BOOL) add: anObject;

- (id *) replace: anObject;

- (BOOL) contains: aKey;

- (BOOL) containsKey: aKey;

- at: aKey;

- (BOOL)at: aKey memberSlot: (id **)memptr;


- remove: aKey;

- removeKey: aKey;

- (void)forEach: (SEL)aSelector;

- (void)forEach: (SEL)aSelector : arg1;

- (void)forEach: (SEL)aSelector : arg1 : arg2;

- (void)forEach: (SEL)aSelector : arg1 : arg2 : arg3;

- (void)forEachKey: (SEL)aSelector;

- (void)forEachKey: (SEL)aSelector : arg1;

- (void)forEachKey: (SEL)aSelector : arg1 : arg2;

- (void)forEachKey: (SEL)aSelector : arg1 : arg2 : arg3;


- (void)describeForEach: outputCharStream;

- (void)describeForEachID: outputCharStream;

- begin: aZone;



- createIndex: (id <Zone>)aZone fromMember: anObject ;

- (void) drop;

- (void) removeAll;

- (void) deleteAll;


//New thing I made up because I didn't like the Swarm's Set default behavior
- forcefulReplace: anObject;

-(void)  walk: (void (*)(void*,void*)) aFunction Param: (void *) PARAM;

- traverse: (avl_traverser *) travPtr;
- (avl_tree *) getTree;
- atOffset: (unsigned) x;


- getFirst;
- getLast;

- removeFirst;
- removeLast;


- (void)setInitialValue: initialValue;

- beginPermuted: aZone;
- (BOOL) allSameClass;
- (void) setIndexFromMemberLoc: (int)byteOffset ;


@end


@interface AVLSetIndex_c: Index_any <KeyedCollectionIndex>
{

  int offset;
  avl_traverser trav;
  id currentObject;
@public
   avl_tree * tree; 
}
- init;
- next;
- prev;
- get;
- replace: anObject;
- remove;
- (id <Symbol>)getLoc;
- (void)setLoc: (id <Symbol>)locSymbol;
- (int)getOffset;
- setOffset: (unsigned)targetOffset;

- getCollection;

//- (void)mapAllocations: (mapalloc_t)mapalloc;
- put: anObject;
-_I_Using;

@end















