// Citizen.m


// Paul Johnson <pauljohn@ku.edu> 2002
// This code is made available under the GPL (GNU general public license)
// with the additional proviso that any users notify the author of their
// use and share their code with the original author.  A copy of version
// 2 of the GPL is included in this directory, but you can use a newer
// GPL at your choice.  The author assumes no liability for any 
// inconvenience or damage caused by this software and offers no 
// warranty of any kind.

#import "Citizen.h"
#import <random.h>
#import <math.h>

#define discrete2dAt(l, x, y) ((l) + (gridOffsets)[(y)] + (x))

@implementation Citizen


- createEnd
{
  if (positionGrid == nil || opinionGrid == nil)
    [InvalidCombination
      raiseEvent: "Agent was created without a world \n"];
  worldXSize = [positionGrid getSizeX];
  worldYSize = [positionGrid getSizeY];
  periodsOpinion=0;
  [self drawPandS];
  return self;
}

- (void)setIDNumber: (int) anumber
{
  idnum = anumber;
}


-(void) drawPandS
{
  persuas = [ myUniformDbl getDoubleWithMin: 0L withMax: 100L ];
  support = [ myUniformDbl getDoubleWithMin: 0L withMax: 100L ] ;
  // persuas = 100; 
  //support = 100;
  // persuas = (double) [myUniformInt getIntegerWithMin: 0L withMax: 100L];
  //support = (double) [myUniformInt getIntegerWithMin: 0L withMax: 100L];
}

- setPositionGrid: (MultiGrid2d *)w OpinionGrid: (LataneGrid *)h
{
  if (positionGrid != nil || opinionGrid != nil)
    {
      [InvalidArgument
        raiseEvent:
          "You can only set the world/heat of a citizen at creation time\n"];
    }
  positionGrid = w;
  opinionGrid = h;

  return self;
}


- (void)setUniformDoubleDist: aDist
{
  myUniformDbl = aDist;
}

- (void) setUniformIntegerDist: aDist
{
  myUniformInt = aDist;
}

-(void) setType: (int) aType
{
  citizenType = aType;
}

- setMovement: (BOOL)mov
{
     movement=mov;
     return self;
}

- setX: (int)inX Y: (int)inY
{
  x = inX;
  y = inY;
  xlag = x;
  ylag = y; 
 
  [positionGrid add: self atX: x Y: y];
  // [opinionGrid addPersonAtX: x Y: y];
  
  return self;
}

- (void)setPersuas: (double)input
{
  persuas= input;
}


- (double)getPersuas
{
  return persuas;
}


- (void)setSupport: (double)input
{
  support = input;
}


- (double)getSupport
{
  return support;
}


- (void)setInitialOpinion: (int)xin
{
  opinion = xin;
  [opinionGrid addOpinion: xin Persuasion: persuas Support: support AtX: x Y: y Type: citizenType ];
  opinionlag1 = xin;

}


//pj-2001-05-02 ?????is this right--no need to lag it.

- (int)getOpinion
{
  return opinion;
}

- (int)getMinority
{
  return bolsteredMinority;
}

- (int)getChanged
{
   return changedOpinion;
}

- (double) getPressureHere
{
   return pressureHere;
}


- updateBeliefs
{
  int npersuasion=0, nsupport=0;
  double persuasionAgainstMe = [opinionGrid getPersuasionFor: (1-opinion) InVicinityX: x Y: y N: &npersuasion];
  double supportForMe = [opinionGrid getSupportFor: opinion  InVicinityX: x Y: y N: &nsupport];

  pressureHere = 0;
  
  if (npersuasion>0) persuasionAgainstMe /= sqrt(npersuasion);
  if (nsupport>0) supportForMe /= sqrt(nsupport);
  //if ((nsupport>npersuasion && persuasionAgainstMe>supportForMe) ||(nsupport<npersuasion && persuasionAgainstMe<supportForMe))
  // printf("\n id=%d, UpdateBeliefs: npersuasion=%d persuasionagainst=%f nsupport=%d support=%f\n",idnum, npersuasion,persuasionAgainstMe,nsupport,supportForMe);
  pressureHere= 50.0 + persuasionAgainstMe - supportForMe;
  //if (pressureHere>50) 
  //printf("pressureHere = %f",pressureHere);
  
  if (nsupport < npersuasion && persuasionAgainstMe < supportForMe) bolsteredMinority = 1; 
  else bolsteredMinority=0;

  return self;
}


- considerOpinion
{
  raiseEvent(WarningMessage,"subclass must implement consider opinion");
  return self; 
}

- (int)changeOpinion
{
  [opinionGrid removeOpinion: opinion Persuasion: persuas Support: support AtX: x Y: y Type: citizenType];

  opinionlag1 = opinion;
  opinion = 1 - opinion;
  if (opinion != 0 && opinion != 1) printf("sucker: %d", opinion);
  [self drawPandS]; 
  [opinionGrid addOpinion: opinion Persuasion: persuas Support: support   AtX: x Y: y  Type: citizenType ]; 
  // printf("opinion changed to %d, was %d\n",opinion,opinionlag1);
  changedOpinion = 1;
  periodsOpinion = 0;
 
  return opinion;
}


-step
{ 
  [self updateBeliefs];  
  [self considerOpinion];
  return self;
}


//must update pressureHere to equal current situation after moving.

-(BOOL) scoutLocationX: (int *) newx Y: (int *) newy Discontent: (double*) newDiscontent 
{
  BOOL foundNewLocation = NO;
  //  printf("scout location %d %d %d %d \n",x,y, *newx, *newy);
  if ( opinion == 0)
    {
      if (pressureHere >0)
	{
	  //go to place with lower discontent than here
	  * newDiscontent = [self findExtremeType: safe X: newx Y: newy];
	  if (*newDiscontent < pressureHere)
	    {
	      foundNewLocation = YES;
	    }
	}
    }
  else 
    {
      if ( pressureHere < 100)
	//go to  place with higher discontent than here
	{ 
	  double temp;
	  temp = [self findExtremeType: 
				       riotous X: newx Y: newy];
	  
	  *newDiscontent = temp;
	  if (*newDiscontent > pressureHere)
	    {
	      foundNewLocation = YES;
	    }	    
	}
    }

  return foundNewLocation;
}



- (double) findExtremeType: (ExtremeType) type X: (int *) px Y: (int *) py
{
  int i;
  double  bestValue;

  int offset = 0;
  int xpoints[] = {0,0,0,0,0,0,0,0};
  int ypoints[] = {0,0,0,0,0,0,0,0};
  int xVals[8] = {*px-1, *px-1, *px-1, *px  , *px  , *px+1, *px+1, *px+1};
  int yVals[8] = {*py-1, *py  , *py+1, *py-1, *py+1, *py-1, *py, *py+1};
  int count = 0;

  //suspected multiple lookups were slow, so now I grab the whole lattice once an repeatedly access it.

  long *visiblePop0 = [opinionGrid getOpinionPointer: 0];
  long *visiblePop1 = [opinionGrid getOpinionPointer: 1];
  //   long *visiblePopulation = [opinionGrid getPopPointer];
  long *gridOffsets =  [opinionGrid getGridOffsets];

  if ( *px == 0 )
    {
     int correctx = worldXSize-1;
      xVals[0]= correctx;
      xVals[1]= correctx;
      xVals[2]= correctx;
    }
  
  else if ( *px == worldXSize- 1 )
    {
      int correctx = 0;
      xVals[5]= correctx;
      xVals[6]= correctx;
      xVals[7]= correctx;
    }

  if ( *py == 0 )
    {
      int correcty = worldYSize-1;
      yVals[0]= correcty;
      yVals[3]= correcty;
      yVals[5]= correcty;
    }
  
  else if ( *py == worldYSize- 1 )
    {
     int correcty = 0;
      yVals[2]= correcty;
      yVals[4]= correcty;
      yVals[7]= correcty;
    }
  

  bestValue = pressureHere;


//    printf("x=%d, y=%d, protest is %ld pop is %ld \n",x,y, *visibleProtestFrom(x,y), *visiblePopulationFrom(x,y));
//    printf("grid says number here is %d \n", [positionGrid  numberOfObjectsAtX: x Y: y]);
//    {
//      int nn,pp;
//      nn=[opinionGrid getOpinionInVicinityX: x Y: y N: &pp];
    
//      printf("whereas the direct return is %d of %d\n",nn,pp);
//    }

  //in fET:
  //now step through the 8 neighbors:
  for (i = 0; i < 8 ; i++) 
    {  
      double valueHere = 0;
      double numYesExclMe = 0;
      int numVisiblePeopleExclMe = 0;
  
    numYesExclMe =*discrete2dAt(visiblePop1,xVals[i],yVals[i]) - opinionlag1;
    numVisiblePeopleExclMe = *discrete2dAt(visiblePop0,xVals[i],yVals[i]) + *discrete2dAt(visiblePop1,xVals[i],yVals[i]) - 1;
	  //  printf ("Async citizen.m numVisiblePeopleExclMe=%d\n",numVisiblePeopleExclMe);

	  // DB 2001-08-18.
          // Self-Awareness Issue5: Note that in synchronous, the printf 
	  // below will print 8*pop lines of -1's for numVisPeopleExclMe on 
	  // the first iteration, since the existence of everyone is not 
	  // flushed through the world until after everyone has looked 
	  // around once and seen his own nonexistence.  On that first 
	  // iteration, he can't see himself but he still subtracts 
	  // himself.  That does not cause problems though, because
	  // the negative numVisPeopleExclMe never gets used, since the 
	  // valueHere maker (which uses numVisPeopleExclMe) only runs 
	  // when numVisProtExclMe > 0.  Otherwise valueHere is zero.
	
	  if (numVisiblePeopleExclMe < numYesExclMe)
	      printf("%d,%d, I see numYesExclMe= %f raw numVisPeopleExclMe= %d\n", xVals[i],yVals[i] , numYesExclMe, numVisiblePeopleExclMe);

	  //pj 2001-06-15. Must not subtract 1 from denominator here
	  //if we already removed one above!

	  // DB 2001-08-18. Self-Awareness Issue6: We had it this way: 
          // if ( (numVisProt > 0) && numVisiblePeople-1 != 0)
	  // On Aug.18, 2001, Dave changed it to this: 
        
      if ( (numYesExclMe > 0) && numVisiblePeopleExclMe != 0)
	{
	  valueHere=100.0*numYesExclMe/(numVisiblePeopleExclMe);
  printf ("numVisiblePeopleExclMe is %d, valueHere is %f\n", numVisiblePeopleExclMe, valueHere);
	}

	     
      printf("%d,%d, findXtype: numVisProtExclMe= %f numVisPeopleExclMe= %d valueHere=%f\n", xVals[i],yVals[i] , numYesExclMe, numVisiblePeopleExclMe, valueHere);
           

      if ( valueHere == bestValue ) 
	{
	  xpoints[count] = xVals[i];
	  ypoints[count] = yVals[i];
	  count++;
	}
      else if(type==safe && valueHere < bestValue)
	{ 
	  bestValue = valueHere; 
	  xpoints[0] = xVals[i];
	  ypoints[0] = yVals[i];
	  count = 1;
	}
      else if(type==riotous && valueHere > bestValue)
	{ 
	  bestValue = valueHere; 
	  xpoints[0] = xVals[i];
	  ypoints[0] = yVals[i];
	  count = 1;
	}
    }

  //if we have more than one equally good alternative, choose randomly among them
  if (count > 1)
    offset = [uniformIntRand getIntegerWithMin: 0L 
                           withMax: count - 1];
  // choose a point at random from the heat list
  //  bestCell = [safetyList atOffset: offset];

  *px = xpoints[offset];
  *py = ypoints[offset];
  
  return bestValue;
}

- (void)sayHi
{

  printf("x: %d y: %d\n",x,y);
  
}
@end








