// LataneGrid.m

// Paul Johnson <pauljohn@ku.edu> 2002
// This code is made available under the GPL (GNU general public license)
// with the additional proviso that any users notify the author of their
// use and share their code with the original author.  A copy of version
// 2 of the GPL is included in this directory, but you can use a newer
// GPL at your choice.  The author assumes no liability for any 
// inconvenience or damage caused by this software and offers no 
// warranty of any kind.

#import "MyArguments.h"
#import <random.h>
#import "LataneGrid.h"
#import "AppSpecificCell.h"
#import "Citizen.h"
#import <math.h>


#define D2D(l,off,x,y) ((l) + (off)[(y)] + (x))

@implementation LataneGrid


+ create: aZone setSizeX: (unsigned)x Y: (unsigned)y
{
  LataneGrid *obj = [self createBegin: aZone];
  obj->xsize = x;
  obj->ysize = y;
 
  return [obj createEnd];
}



- createEnd
{
  unsigned 
    int i,j = 0;
  
  //retrieve parameters (use either getInt or get method, your choice)
  wrapAround = [(id)arguments getWrapAround];
  SYNCHRONOUS = [(id)arguments getSYNCHRONOUS];
  radius = getInt(arguments, "visionRadius");

  if ( wrapAround == YES )
    {
     int smallside = (xsize < ysize) ? xsize: ysize;
     if (smallside % 2 == 0) //small side is even number of cells
       {
	 if (radius >= smallside/2) radius = (smallside/2) -1;
       }
     else //smallside is odd number of cells
       {
	 if (radius > smallside/2) radius = (smallside/2);
       }
     // fprintf(stderr, "Smallside %d Smallside/2 %d radius %d", smallside, smallside/2, radius);
    }
  
  correction = 1.0/radius*radius;

  if ( wrapAround == NO )
    {
      int smallside = (xsize < ysize) ? xsize: ysize;
      if (radius > smallside-1) radius = smallside-1; 
      // fprintf(stderr,"Smallside %d Smallside/2 %d radius %d", smallside, smallside/2, radius);
    }


  accumP = [[self getZone] allocBlock: (OPINIONS*sizeof(double*))];
  accumS = [[self getZone] allocBlock: (OPINIONS*sizeof(double*))];

  for (i =0; i < OPINIONS; i++)
    {
      accumP[i] = [[self getZone] allocBlock: (xsize * ysize * sizeof(double))];
      accumS[i] = [[self getZone] allocBlock: (xsize * ysize * sizeof(double))];
      memset (accumP[i], 0, xsize * ysize * sizeof (double));
      memset (accumS[i], 0, xsize * ysize * sizeof (double)); 
    }

  gridOffsets = [[self getZone] alloc: ysize * sizeof (*offsets)];

  for (i = 0; i < ysize; i++)
    gridOffsets[i] = xsize * i; 

  visibleP = [[self getZone] allocBlock: (OPINIONS*sizeof(double *))];
  visibleS = [[self getZone] allocBlock: (OPINIONS*sizeof(double *))];

  for (i =0; i < OPINIONS; i++)
    {
      visibleP[i] =[[self getZone] allocBlock: (xsize * ysize * sizeof(double))];
      visibleS[i] =[[self getZone] allocBlock: (xsize * ysize * sizeof(double))];
      memset (visibleP[i], 0, xsize * ysize * sizeof (double));
      memset (visibleS[i], 0, xsize * ysize * sizeof (double)); 
    }

  visibleOpinion = [[self getZone] allocBlock: (OPINIONS*sizeof(long *))];
  for (i =0; i < OPINIONS; i++)
    {
      visibleOpinion[i] = [[self getZone] allocBlock: (xsize * ysize * sizeof(long))];
      memset (visibleOpinion[i], 0, xsize * ysize * sizeof (long)); 
    }

  popChange = [[self getZone] allocBlock: (OPINIONS*sizeof(long *))];
  for (i =0; i < OPINIONS; i++)
    {
      popChange[i] = [[self getZone] allocBlock: (xsize * ysize * sizeof(long))];
      memset (popChange[i], 0, xsize * ysize * sizeof (long)); 
    }
  
  for (i = 0; i < NUMTYPES; i++)
    {
      for (j = 0; j < OPINIONS; j++)
	{
	  num[j][i]=0;
	}
      numType[i]=0;
    }

  return [super createEnd];
}


- setNumPPL:(int)n
{
  numPPL=n;
  return self;
}

- setNumType: (int)b To: (int)number
{
  numType[b] = number ;
  return self;
}

- (double)getPct
{
  int i, total=0;
  double z=0;
  

  for (i = 0; i < NUMTYPES; i++)
    total += num[1][i];
  
  if(numPPL != 0)
    {
      z = (double)total/numPPL;
    }

  //printf("NHOOD Figures %d %d \n", numSupporters, numPPL);
  return z;
}

- (double)getPctType: (unsigned)aType
{
  
  if ( numType[aType] != 0 )
    return (double) num[1][aType]/numType[aType];
  return 0;
}

- (double)getPctTypeOne
{
  // [self printDiagnostics];
  if ( numType[1] != 0 )
    return (double) num[1][1]/numType[1];
  return 0;
}



//pj-2001-05-02 The stepRule only needs to be called for a SYNCHRONOUS model,
//because that one accumulates the changes and pushes them all through at once.
//The ASYNCHRONOUS alternative flushes through each agent's action immediately.

- stepRule
{
  unsigned xx, yy;
  for (xx = 0; xx < xsize; xx++)
    for (yy = 0; yy < ysize; yy++)
      {
	int i, intlevel=0;
	double dbllevel =0;
	//   fprintf(stderr,"step Rule (%d,%d) accumulator= %d \n",xx,yy,*accumulatorAt(xx,yy));

	//first flush through the population changes

	if (wrapAround == YES) 
	  {
	    for ( i = 0; i < OPINIONS; i++ )
	      {
	      if ( (intlevel=*D2D(popChange[i],gridOffsets,xx,yy) ) != 0)
		{
		  [self updateOpinion: i Impact: intlevel X: xx Y: yy ];
		}
	      }
	  }
	else
	  {
	    for ( i = 0; i < OPINIONS; i++ ) 
	      {
		if ( (intlevel=*D2D(popChange[i],gridOffsets,xx,yy)) != 0)
		  {
		    [self updateOpinionNoWrap:i Impact: intlevel X: xx Y: yy];
		  }
	      }
	  }
	
	for ( i = 0; i < OPINIONS; i++)
	  {
	    if ( (dbllevel = *D2D(accumP[i],gridOffsets,xx,yy) ) != 0 )
	      {
		if (wrapAround == NO)
		  //[self updateVisPNoWrap: i Impact: dbllevel X: xx Y: yy];
		  [self updateNoWrapDblPtr: &visibleP[i][0] Impact: dbllevel X: xx Y: yy];
		else
		  // [self updateVisP: i Impact: dbllevel X: xx Y: yy];
		  [self updateDblPtr: &visibleP[i][0] Impact: dbllevel X: xx Y: yy];	     
	      }
	    if ( (dbllevel = *D2D(accumS[i],gridOffsets,xx,yy) ) != 0 )
	      {
		if (wrapAround == NO)
		  // [self updateVisSNoWrap: i Impact: dbllevel X: xx Y: yy];
		  [self updateNoWrapDblPtr: &visibleS[i][0] Impact: dbllevel X: xx Y: yy];
		else
		  //[self updateVisS: i Impact: dbllevel X: xx Y: yy];
		  [self updateDblPtr: &visibleS[i][0] Impact: dbllevel X: xx Y: yy];
	      }
	    //clear off the accumulators, since they have flushed through.
	    //tried equivalent of "fastFillWithValue: to erase
	    //these, but it gave no speedup
	    *D2D(popChange[i], gridOffsets, xx, yy) = 0;
	    *D2D(accumS[i],gridOffsets, xx, yy) = 0;
 	    *D2D(accumP[i],gridOffsets, xx, yy) = 0;
	  }
      }
  
  for (xx = 0; xx < xsize; xx++)
    {
      for (yy = 0; yy < ysize; yy++)
	{	
	  double pct = 9999.0;
	  double a = *D2D(visibleOpinion[1],gridOffsets,xx,yy);
	  double b = *D2D(visibleOpinion[0],gridOffsets,xx,yy);
	  double c=0,d=0;
	  
	  if (a>0) c= (*D2D(visibleP[1],gridOffsets,xx,yy) / (double) sqrt(a));
	  if (b>0) d= (*D2D(visibleP[0],gridOffsets,xx,yy) / (double) sqrt(b));
	  
	  //pct = 100.0 / (1.0 + exp(c-d)); //logit transform to normalize to 0-100 scale
	  pct = 100.0 /(1.0 + exp( d-c));
	  *(discrete2dSiteAt(lattice, offsets, xx, yy)) = (id)((long)(floor(pct)));//round down
	  // printf("latane grid double value = %f,P1=%f, NP1=%f, avgP1=%f, P0=%f, avgP0=%f, NP0=%f, S0=%f S1=%f \n",pct, 
	  //  	       *D2D(visibleP[0],gridOffsets,xx,yy),a,c, 
	  //  	       *D2D(visibleP[1],gridOffsets,xx,yy),b,d,
	  //  	       *D2D(visibleS[0],gridOffsets,xx,yy),
	  //  	       *D2D(visibleS[1],gridOffsets,xx,yy));
	}
    }
  return self;
}



//For SYNCHRONOUS updating, thhis method is called when the world
//wraps around.  IT takes the stance and adds its "impact" (reflecting
//changes by a number of agents, possibly) across the whole
//neighborhood which can see this cell. If the Impact is negative, it
//means people have changed away from that stance and so they are
//removed from the visible opinion pattern

- (void)updateOpinion: (int)stance Impact: (int)input X: (int)x Y: (int)y
{
  int r, c;
  int limit = 2*radius + 1;  // vision range is 2*radius+1
  int xcoord=0, ycoord=0;
  for(r=0; r < limit; r++)
    {
      xcoord = x + r - radius; //start at far left
      if ( (xcoord < 0)||(xcoord > (int)xsize-1) ); // outOfBoundsX
      {
	xcoord = (xcoord + xsize) % xsize;
      }
     for(c=0; c < limit ;c++)
	{
	  ycoord= y + c - radius; //start at far low y
	  
	  if ((ycoord < 0)||(ycoord > (int)ysize-1))   //outOfBoundsY 
	    {
	      ycoord = (ycoord + ysize) % ysize;
	    }
	  // fprintf(stderr,"Update Opinion r,c = (%d,%d)x= %d, y= %d,xcoord,ycoord=(%d,%d)  input= %d \n",r,c, x,y,xcoord,ycoord,input);
	  *D2D(visibleOpinion[stance],gridOffsets,xcoord,ycoord) += input;
	}
    }
}

//Same as updateOpinion:Impact:X:Y: except that the impact does not
//wrap around the edges.

- (void) updateOpinionNoWrap: (int) stance Impact: ( int) input X: (int) x Y: (int) y
{
  int xcoord=0, ycoord=0;
  int leftX = x - radius;
  int rightX = x + radius + 1; //one beyond included region
  
  int lowY = y - radius;
  int hiY = y + radius + 1; //one beyond included region

  if (lowY < 0) lowY = 0;
  if (hiY > (int)ysize) hiY = ysize;

  if (leftX < 0) leftX = 0;
  if (rightX > (int)xsize ) rightX = xsize;

  for(xcoord = leftX; xcoord < rightX; xcoord++)
    {
      for(ycoord = lowY; ycoord < hiY; ycoord++)
	{
	  // fprintf(stderr,"Update Opinion x= %d, y= %d,xcoord,ycoord=(%d,%d)  input= %d \n",x,y,xcoord,ycoord,input);
	  *D2D(visibleOpinion[stance],gridOffsets,xcoord,ycoord) += input;
	}
    }
}


// In case you distrust the abstract multipupose updateDblPtr: and
// updateNoWrapDblPtr: methods below, you can remove the comments here
// and verify the numbers.

// - (void) updateVisP: (int)stance Impact: (double) input X: (int) x Y: (int) y
// {
//   int r, c, xcoord, ycoord;
//   for(r=0;r<(2*radius)+1;r++)
//     {
//       xcoord = x+r-radius;
//       if ( xcoord < 0 || xcoord > xsize-1  )
// 	{
// 	  xcoord = (xcoord + xsize) % xsize;
// 	}
//       for(c=0;c<(2*radius)+1;c++)
// 	{
// 	  double dist = 1.0;
// 	  ycoord= y+c-radius;
// 	  if (ycoord < 0 || ycoord > ysize-1) ycoord = (ycoord + ysize) % ysize;
// 	  dist = 1.0 + sqrt((c-radius)*(c-radius) +(r-radius)*(r-radius));
// 	  *D2D(visibleP[stance],gridOffsets,xcoord,ycoord) += input/(dist*dist);
// 	}
//     }
// }



// - (void) updateVisPNoWrap: (int)stance Impact: (double) input X: (int) x Y: (int) y
// {
//   int xcoord=0, ycoord=0;
//   int leftX = x - radius;
//   int rightX = x + radius;
  
//   int lowY = y - radius;
//   int hiY = y + radius;

//   if (lowY < 0) lowY = 0;
//   if (hiY > ysize-1) hiY = ysize -1;

//   if (leftX < 0) leftX = 0;
//   if (rightX > xsize -1) rightX = xsize-1;

//   for(xcoord = leftX; xcoord < rightX +1; xcoord++)
//     {
//       for(ycoord = lowY; ycoord < hiY+1 ; ycoord++)
// 	{
// 	  double dist = 1.0 + sqrt((x-xcoord)*(x-xcoord) +(y-ycoord)*(y-ycoord));
// 	  *D2D(visibleP[stance],gridOffsets,xcoord,ycoord) += input/(dist*dist);
// 	}
//     }
// }




// - (void) updateVisS: (int)stance Impact: (double) input X: (int) x Y: (int) y
// {
//   int r, c, xcoord, ycoord;
//   for(r=0;r<(2*radius)+1;r++)
//     {
//       xcoord = x+r-radius;
//       if ( xcoord < 0 || xcoord > xsize-1  )
// 	{
// 	  xcoord = (xcoord + xsize) % xsize;
// 	}
//       for(c=0;c<(2*radius)+1;c++)
// 	{
// 	  double dist = 1.0;
// 	  ycoord= y+c-radius;
// 	  if (ycoord < 0 ||  ycoord > ysize-1) ycoord = (ycoord + ysize) % ysize;
// 	  dist = 1.0 + sqrt((c-radius)*(c-radius) +(r-radius)*(r-radius));
// 	  *D2D(visibleS[stance],gridOffsets,xcoord,ycoord) += input/(dist*dist);
// 	}
//     }
// }



// - (void) updateVisSNoWrap: (int)stance Impact: (double) input X: (int) x Y: (int) y
// {
//   int xcoord=0, ycoord=0;
//   int leftX = x - radius;
//   int rightX = x + radius;
  
//   int lowY = y - radius;
//   int hiY = y + radius;

//   if (lowY < 0) lowY = 0;
//   if (hiY > ysize-1) hiY = ysize -1;

//   if (leftX < 0) leftX = 0;
//   if (rightX > xsize -1) rightX = xsize-1;

//   for(xcoord = leftX; xcoord < rightX +1; xcoord++)
//     {
//       for(ycoord = lowY; ycoord < hiY+1 ; ycoord++)
// 	{
// 	  double dist = 1.0 + sqrt((x-xcoord)*(x-xcoord) +(y-ycoord)*(y-ycoord));
// 	  *D2D(visibleS[stance],gridOffsets,xcoord,ycoord) += input/(dist*dist);
// 	}
//     }
// }


//general purpose method to distribute a double-valued variable's "input"
//across a square of size "radius" that eminates from x,y. 

- (void)updateDblPtr: (double*)p Impact: (double)input X: (int)x Y: (int)y
{
  int r, c;
  int xcoord, ycoord;
  int limit = 2*radius + 1;
  for ( r=0; r < limit; r++ )
    {
      int diffx = r-radius;
      xcoord = x+diffx;
      if ( xcoord < 0 || xcoord > (int)xsize-1  )
	{
	  xcoord = (xcoord + xsize) % xsize;
	}
      for ( c=0; c < limit; c++ )
	{
	  // try "radial" distance, not Euclidean
	  // int tempx = 0;
	  // int tempy = 0;
	  
	  double dist = 1;
	  int diffy = c-radius;
	  ycoord = y + diffy;
	  if ( ycoord < 0 || ycoord > (int)ysize-1 ) ycoord = (ycoord + (int)ysize) % ysize;
	  //This was the euclidean calculation:
	  dist = 1.0 + sqrt((c-radius)*(c-radius) +(r-radius)*(r-radius));
	  //Here's the flat radial calculation:
	  //dist += abs(diffx)<abs(diffy)?abs(diffy):abs(diffx);
	  *D2D(p,gridOffsets,xcoord,ycoord) += input/(dist*dist); 
	 
	  // 	  printf ("xy=%d,%d coords=%d,%diffx=%d, diffy=%d, dist=%f, input=%f, inp/dist= %f gridval=%f \n",
	  // 	x, y, xcoord, ycoord, diffx, diffy, dist,
	  // 	input, input/dist, *D2D(p,gridO
	  //	 ffsets,xcoord,ycoord));
	}
    }
}


//influence eminates out from x,y, but now it stops at the edge of the grid.
- (void) updateNoWrapDblPtr: (double*) p  Impact: (double) input X: (int) x Y: (int) y
{
  int xcoord=0, ycoord=0;
  int leftX = x - radius;
  int rightX = x + radius +1; //one beyond radius
  
  int lowY = y - radius;
  int hiY = y + radius +1; //one beyond radius

  if (lowY < 0) lowY = 0;
  if (hiY > (int)ysize) hiY = ysize;

  if (leftX < 0) leftX = 0;
  if (rightX > (int)xsize) rightX = xsize;

  for(xcoord = leftX; xcoord < rightX; xcoord++)
    {
      for(ycoord = lowY; ycoord < hiY ; ycoord++)
	{
	  //commented-out lines do distance on "radial" metric
	  //int tempx = fabs(x-xcoord);
	  //int tempy = fabs(y-ycoord);
	  
	  //The radial calculation: 
	  //double dist = 1;
	  //dist += tempx<tempy?tempy:tempx;
	  //Euclidean calculation
	  double dist = 1.0 + sqrt((x-xcoord)*(x-xcoord) +(y-ycoord)*(y-ycoord));
 
	  //dist += tempx > tempy? tempx: tempy;
	  *D2D(p,gridOffsets,xcoord,ycoord) += input/(dist*dist); 
	  // 	  printf ("xy=%d,%d coords=%d,%dtempx=%d, tempy=%d, dist=%f, input=%f, inp/dist= %f gridval=%f \n",
	  // 		 x, y, xcoord, ycoord, tempx, tempy, dist,
	  // 		 input, input/dist, *D2D(p,gridOffsets,xcoord,ycoord));
	}
    }
}



- (int)getVisibleNOpinion: (int)op InVicinityX: (int)x Y: (int)y
{
  return *(D2D(visibleOpinion[op],gridOffsets,x,y));
}


- (double)getSupportFor: (int)stance InVicinityX: (int)x Y: (int)y N:(int*)n_ptr
{
  *n_ptr = *D2D(visibleOpinion[stance],gridOffsets,x,y);
  return  *D2D(visibleS[stance],gridOffsets,x,y);
}  

- (double)getPersuasionFor: (int)stance InVicinityX: (int)x Y: (int)y N: (int*)n_ptr
{
  *n_ptr = *D2D(visibleOpinion[stance],gridOffsets,x,y);
  return *D2D(visibleP[stance],gridOffsets,x,y);
}


- (double*)getPersuasionPointer: (int)op
{
  return visibleP[op];
}



- (double*)getSupportPointer: (int)op
{
  return visibleS[op];
}



- (long*)getOpinionPointer: (int)op
{
  return visibleOpinion[op];
}


- (long *)getGridOffsets
{
  return gridOffsets;
}

/*"This is how citizens register their opinion in this model. They add an opinion which
has value 'stance' and then they must record their own persuasive and supportive power
as it eminates out from x,y. "*/
- (long)addOpinion: (int)stance Persuasion: (double)p Support: (double)s AtX: (int)x Y: (int)y Type: (int)aType
{
  if (SYNCHRONOUS)
    {
      *D2D(accumP[stance],gridOffsets,x,y) += p;
      *D2D(accumS[stance],gridOffsets,x,y) += s;
      *D2D(popChange[stance],gridOffsets,x,y) += 1;
    }
  else
    {
      int r, c, xcoord = 0, ycoord = 0;
      int limit = 2*radius + 1;
      double pct;
   
      for (r=0; r<limit; r++)
	{
	  BOOL outOfBoundsX = NO, outOfBoundsY = NO;
	  xcoord = x+r-radius;
	  outOfBoundsX = (xcoord < 0) || (xcoord > (int)xsize -1);
	  if (outOfBoundsX)
	    {
	      if (wrapAround == NO) continue;
	      else xcoord = (xcoord + xsize)%xsize;
	    }
	  for (c=0; c< limit; c++)
	    {
	      double aa=0, bb=0,cc=0,dd=0;
	      double dist = 1.0 + sqrt((c-radius)*(c-radius) +(r-radius)*(r-radius));
	      ycoord=y+c-radius;
	      outOfBoundsY =  (ycoord < 0)||(ycoord > (int)ysize-1);
	      if (outOfBoundsY)
		{
		  if (wrapAround == NO) continue; //if outside bounds, then quit here
		  else ycoord = (ycoord + ysize) % ysize;
		}
	      *D2D(visibleP[stance],gridOffsets,xcoord,ycoord) += p/(dist*dist) ;
	      *D2D(visibleS[stance],gridOffsets,xcoord,ycoord) += s/(dist*dist) ;
	      
	      *D2D(visibleOpinion[stance],gridOffsets,xcoord,ycoord) += 1;
	      
	      aa = *D2D(visibleOpinion[1],gridOffsets,xcoord,ycoord); 
	      bb = *D2D(visibleOpinion[0],gridOffsets,xcoord,ycoord);	  
	      if (aa>0) cc= (*D2D(visibleP[1],gridOffsets,xcoord,ycoord) / (double) sqrt(aa));
	      if (bb>0) dd= (*D2D(visibleP[0],gridOffsets,xcoord,ycoord) / (double) sqrt(bb));
	      
	      pct = 50+ correction*(cc  - dd);
	      
	      *(D2D(lattice, offsets, xcoord, ycoord)) = (id)((long)(floor(pct+0.5)));   
	    }
	}
      
      //fprintf(stderr, "(%d %d)ASynch-Protestor Added: now visibleProtest=%ld,visibleOpinion= %ld\n",x,y, *visibleInfluenceAt(x,y),*visibleOpinionulationAt(x,y)) ;
      
    }
  num[stance][aType]++;
  return (long)*(D2D(lattice, offsets, x, y)) ;    
}

/*"If an agent decides to change its mind, it has to remove the old
opinion it registered and then add its opinion(see previous method).
Because the agent has to remove not only its opinion but also its
impact on others, this method takes additional arguments.  Note that, in
a SYNCHRONOUS updating model of the original Nowak/Latane sort,
this method just adds the opinion to the accumulators, which hold
the impact until it is time to update the whole grid."*/

- (long)removeOpinion: (int)stance Persuasion: (double)p Support: (double)s AtX: (int)x Y: (int)y Type: (int)aType
{
  if (SYNCHRONOUS)
    {
    //  printf("(%d,%d) Remove protestor: accumulator before %d, visProt here before removal: %f, visPop=%ld popChangePending=%ld \n" ,x,y,*accumulatorAt(x,y),*visibleInfluenceAt(x,y),*visibleOpinionulationAt(x,y),*popChangeAt(x,y));
    *D2D(accumP[stance],gridOffsets,x,y) -= p;
    *D2D(accumS[stance],gridOffsets,x,y) -= s;
    *D2D(popChange[stance],gridOffsets,x,y) -= 1;
    }
  else
    {
      
      int r, c, xcoord = 0, ycoord = 0;
      int limit = 2*radius + 1;
      double pct;
      for (r=0; r< limit; r++)
	{
	  BOOL outOfBoundsX = NO, outOfBoundsY = NO;
	  xcoord = x+r-radius;
	  outOfBoundsX = (xcoord < 0) || (xcoord > (int)xsize -1);
	  if (outOfBoundsX)
	    {
	      if (wrapAround == NO) continue;
	      else xcoord = (xcoord + xsize)%xsize;
	    }
	  
	  for (c=0; c< limit; c++)
	    {
	      double aa=0,bb=0,cc=0,dd=0;
	      double dist = 1.0 + sqrt((c-radius)*(c-radius) +(r-radius)*(r-radius));
	      ycoord=y+c-radius;
	      outOfBoundsY = (ycoord < 0)||(ycoord > (int)ysize-1);
	      if (outOfBoundsY)
		{
		  if (wrapAround == NO) continue; //if outside bounds, then quit here
		  else ycoord = (ycoord + ysize) % ysize;
		} 
	      *D2D(visibleP[stance],gridOffsets,xcoord,ycoord) -= p/(dist*dist);
	      *D2D(visibleS[stance],gridOffsets,xcoord,ycoord) -= s/(dist*dist);
	      
	      *D2D(visibleOpinion[stance],gridOffsets,xcoord,ycoord) -= 1;
	      
	      aa = *D2D(visibleOpinion[1],gridOffsets,xcoord,ycoord); 
	      bb = *D2D(visibleOpinion[0],gridOffsets,xcoord,ycoord);	  
	      if (aa>0) cc= (*D2D(visibleP[1],gridOffsets,xcoord,ycoord) / (double) sqrt(aa));
	      if (bb>0) dd= (*D2D(visibleP[0],gridOffsets,xcoord,ycoord) / (double) sqrt(bb));
	      
	      pct = 50+ correction*(cc  - dd);
	      	      
	      *(discrete2dSiteAt(lattice, offsets, xcoord, ycoord)) = (id)((long)(floor(pct+0.5)));   
	    }
	}
    }
  num[stance][aType]--;
  return (long) *(D2D(lattice, offsets, x, y)) ;
}


-(void) printDiagnostics
{
  unsigned int i, j;
 
  for (i = 0; i < xsize; i++)
    {
      for(j=0; j < ysize; j++)
	{
	  fprintf(stderr," %ld,%ld,%f,%f,%f,%f,%ld \n",
		  *D2D(visibleOpinion[0],gridOffsets,i,j),*D2D(visibleOpinion[1],gridOffsets,i,j),
		  *D2D(visibleP[0],gridOffsets,i,j), *D2D(visibleP[1],gridOffsets,i,j),
		  *D2D(visibleS[0],gridOffsets,i,j), *D2D(visibleS[1],gridOffsets,i,j),
		  (long)[self getValueAtX: i Y: j]); 
	}
      fprintf(stderr,"\n");
    }
  fprintf(stderr,"\n");
}



@end

