// ModelSwarm.m

// Paul Johnson <pauljohn@ku.edu> 2002
// This code is made available under the GPL (GNU general public license)
// with the additional proviso that any users notify the author of their
// use and share their code with the original author.  A copy of version
// 2 of the GPL is included in this directory, but you can use a newer
// GPL at your choice.  The author assumes no liability for any 
// inconvenience or damage caused by this software and offers no 
// warranty of any kind.

#import "ModelSwarm.h"
#import <random.h>
#import <space.h>
#import <simtools.h>
#import "AppSpecificCell.h"
#import "MyArguments.h"
#import "Citizen2.h"
#import "Citizen1.h"


@implementation ModelSwarm

- getCitizenList
{
  return citizenList;
}

- getListType: (int)x
{
  return [listTypeArray atOffset: x]; 
}

- (MultiGrid2d *)getPositionGrid
{
  return positionGrid;
}

- (LataneGrid *)getInfluenceGrid
{
  return opinionGrid;
}

- createCitizen1X: (unsigned)x  Y: (unsigned)y 
{
  Citizen1 * aCitizen;
  id newGen = [C2TAUS3gen create: [self getZone] setStateFromSeed: [randomGenerator getUnsignedSample]];
  id newGen2 = [MRG7gen create: [self getZone] setStateFromSeed: [randomGenerator getUnsignedSample]];
  id myUniformInt = [UniformIntegerDist
                  create: self 
                  setGenerator: newGen
                  setIntegerMin: 0
                  setMax: 100];

  id myUniformDbl = [UniformDoubleDist
		    create: self
		    setGenerator: newGen2
		    setDoubleMin: 0
		    setMax: 100];


     
  aCitizen = [Citizen1 createBegin: self];
  [aCitizen  setPositionGrid: positionGrid OpinionGrid: opinionGrid];
  [aCitizen setType: 1];
  [aCitizen setUniformIntegerDist: myUniformInt];
  [aCitizen setUniformDoubleDist: myUniformDbl];
  aCitizen = [aCitizen createEnd];

  [aCitizen setX:  x  Y: y ];
  if ( [uniformDblRand getDoubleWithMin: 0L withMax: 1L] > probYES )
    [aCitizen setInitialOpinion: 0];
  else
    [aCitizen setInitialOpinion: 1];

  if (citizensMove == 0)     
    {
      [aCitizen setMovement: NO];
    }
  else
    {
      [aCitizen setMovement: YES];
    }
 
  return  aCitizen;
}



- createCitizen2X:(unsigned)x Y:(unsigned)y
{
  Citizen2 * aCitizen;

  id newGen = [C2TAUS3gen create: [self getZone] setStateFromSeed: [uniformUnsRand getUnsignedSample]];
  id newGen2 = [MRG7gen create: [self getZone] setStateFromSeed: [uniformUnsRand getUnsignedSample]];
  id myUniformInt = [UniformIntegerDist
                  create: self 
                  setGenerator: newGen
                  setIntegerMin: 0
                  setMax: 100];

  id myUniformDbl = [UniformDoubleDist
		    create: self
		    setGenerator: newGen2
		    setDoubleMin: 0
		    setMax: 100];


     
  aCitizen = [Citizen2 createBegin: self];
  [aCitizen  setPositionGrid: positionGrid OpinionGrid: opinionGrid];
  [aCitizen setType: 0];
  [aCitizen setUniformIntegerDist: myUniformInt];
  [aCitizen setUniformDoubleDist: myUniformDbl];
  aCitizen = [aCitizen createEnd];
  [aCitizen setX: x  Y: y ];     
  if (citizensMove == 0)     
    {
      [aCitizen setMovement:  NO];
    }
  else
    {
      [aCitizen setMovement: YES];
    }
  return  aCitizen;
}


+ createBegin: aZone
{
  ModelSwarm *obj;
  id <ProbeMap> probeMap;
  probeMap = [EmptyProbeMap createBegin: aZone];
  [probeMap setProbedClass: [self class]];
  probeMap = [probeMap createEnd];

   obj = [super createBegin: aZone];

  
   obj->worldXSize = getInt(arguments, "worldXSize");
   obj->worldYSize = getInt(arguments, "worldYSize");
   obj->citizensMove = getInt(arguments, "citizensMove");
   obj->randomSeed = getInt (arguments,"randomSeed");
   obj->numPTwo = getInt (arguments, "numPTwo");
   obj->probYES = getDouble (arguments, "probYES");
   obj->SYNCHRONOUS=[(id)arguments getSYNCHRONOUS];
   obj->RANDOMIZED = [(id)arguments getRANDOMIZED];

   //Here is the place for message probes, if you need runtime
   //jolts to the model   
   //  [probeMap addProbe: [probeLibrary getProbeForMessage:
   //                                       "pulseFeature:Percentage:Magnitude:"
   //                                     inClass: [self class]]];
   
   //    [probeMap addProbe: [probeLibrary getProbeForMessage:
   //                                        "pulseFeature:Percentage:ChangeTo:"
   //                                     inClass: [self class]] ];
   [probeLibrary setProbeMap: probeMap For: [self class]];

 
   return obj;
}

//do some checking on parameter settings. Why? Because I'm virtuous:)

- createEnd
{
  if ((citizensMove != 0) && (citizensMove != 1))
    [InvalidCombination
      raiseEvent: "\ncitizensMove has to be 0 or 1.\n"];
  if (numPTwo > numPPL)        
    [InvalidCombination
      raiseEvent: "\numPTwo must be be <= numPPL, friend! \n"];
  return [super createEnd];
}


- initializeRNGs
{
	//following is no longer needed because swarm takes S as a command line argument.
	//Surprisingly, I don't get identical answers if this is uncommented,
	//but expected to. Must be swarm draws some numbers before this call, and this
	//resets the generator.
  //[randomGenerator setStateFromSeed: randomSeed];
//    id newGen = [C2TAUS3gen create: [self getZone] setStateFromSeed: [randomGenerator getInitialSeed]];
//    id newGen2 = [MRG7gen create: [self getZone] setStateFromSeed: [randomGenerator getInitialSeed]];
//    myUniformInt = [UniformIntegerDist
//                   create: self 
//                   setGenerator: newGen
//                   setIntegerMin: 0
//                   setMax: 100];

//    myUniformDbl = [UniformDoubleDist
// 		    create: self
// 		    setGenerator: newGen2
// 		    setDoubleMin: 0
// 		    setMax: 100];
   return self;
}

- buildObjects
{
  int i = 0;
  int j = 0;
  int k = 0;

  [super buildObjects];
  [self initializeRNGs];

  //number of people is equal to the number of cells, at least for starters 
#if 1
  numPPL = worldXSize*worldYSize;
#endif

  positionGrid = [MultiGrid2d create: self setSizeX: worldXSize Y: worldYSize Cells: [AppSpecificCell class]];
  opinionGrid = [LataneGrid create: self setSizeX: worldXSize Y: worldYSize];
  [opinionGrid fastFillWithValue: 0];  
  [opinionGrid setNumPPL: numPPL];

  [opinionGrid setNumType: 0 To: numPTwo];
  [opinionGrid setNumType: 1 To: numPPL-numPTwo];

  // holdingList = [List create: self];
  citizenList = [Array create: self setCount: numPPL];
  
  //pj 2001-4-29: must create listTypeArray

  listTypeArray = [Array create: self setCount: NUMTYPES];

  for (k = 0; k < NUMTYPES; k++)
    {
      id <List> newList = [List create: self];
      [listTypeArray atOffset: k put: newList ];
    }

  //ignore the numPPL parameter, just smoothe people around!
#if  1
  citizenCount = 0;
  for (i = 0; i < worldXSize; i++)
    for (j = 0; j < worldYSize; j++)
    {
        id aCitizen;
	aCitizen = [self createCitizen1X: i Y: j];
	[aCitizen setIDNumber: citizenCount];
	[citizenList atOffset: citizenCount put: aCitizen];
	citizenCount++;
	[[listTypeArray atOffset: 1] addLast: aCitizen];
    }
 
#else
 
  for (i = 0; i < (numPPL - numPTwo); i++)       
    {
      id aCitizen;
      unsigned x,y;
      x=[uniformIntRand
	  getIntegerWithMin: 0L
	  withMax: (worldXSize-1)];
      y=[uniformIntRand getIntegerWithMin: 0L withMax: (worldYSize-1)];
      aCitizen = [self createCitizen1X: x Y: y];
      [aCitizen setIDNumber: citizenCount];
      [citizenList atOffset: citizenCount put: aCitizen];
      citizenCount++;
       [[listTypeArray atOffset: 1] addLast: aCitizen];
    }
#endif

  for (j = 0; j < numPTwo; j++)
    {
      id aCitizen;
            
      aCitizen = [self createCitizen2X: 20 Y: 20];
      [aCitizen setIDNumber: citizenCount];
      [citizenList atOffset: citizenCount put: aCitizen];
      citizenCount++;
      [[listTypeArray atOffset: 0] addLast: aCitizen];
    }

  //  [opinionGrid updatePopulation];  

  return self;
}


-(void) stepThruCitizens 
{
  if (getCurrentTime() > 0)
    {
      id aCitizen;
      
      id index = nil;
      if (SYNCHRONOUS || !RANDOMIZED) index = [citizenList begin: self]; 
      else if (RANDOMIZED) 
 	{
	  id shuffler = [[ListShuffler createBegin: self] createEnd];
	  [shuffler shuffleWholeList: citizenList];
	  [shuffler drop];
	  index = [citizenList begin: self]; // randomly sorted each step.
	}
      for (aCitizen=[index next]; [index getLoc]==Member; aCitizen=[index next])
	[aCitizen step];
      [index drop];
    }
}
  
//        COMMENT: citizens can either look at the grid as it was when
//        last updated (SYNCHRONOUS) or they can do a survey of their
//        neighborhood as it is currently, taking into account the
//        actions of other agents who have already acted in the same
//        turn (ASYNCHRONOUS). The LataneGrid can offer either kind:
  
- buildActions
{
  id modelActions;
  [super buildActions];
  
  modelActions = [ActionGroup create: self]; 

  if (SYNCHRONOUS)
    {
    
      [modelActions createActionTo: opinionGrid message: M(stepRule)];
    }
  
  [modelActions createActionTo: self message: M(stepThruCitizens)];
  //[modelActions createActionTo: opinionGrid message: M(printDiagnostics)]; 

  modelSchedule = [Schedule createBegin: self];
  [modelSchedule setRepeatInterval: 1];
  modelSchedule = [modelSchedule createEnd];
  [modelSchedule at: 0 createAction: modelActions];

  return self;
}


// create files and write headers for variable names
-(void) initializeOutputFiles
{
  sprintf(summaryFileName, "Summary%d", [(id)arguments getRunArg]) ;
  summaryFile = fopen (summaryFileName,"w+");
}


// in case there is any end time info worth saving,use this method to
// do it. To add stuff, use fprint, as in: 
// fprintf ( outputFile,"%ld ",timenow ); 
// fprintf ( outputFile,"%f ", pctProt);

- finishLog
{
  fflush (summaryFile);
  return self;
}


// - chooseSubset
// {
//   [holdingList removeAll];
//   [NSelect select: 30 from: citizenList into: holdingList];
 
//   return holdingList;
// }

- activateIn: swarmContext
{
  // First, activate ourselves via the superclass activateIn: method.
  // Just pass along the context: the activity library does the right thing.

  [super activateIn: swarmContext];

  // Now activate our own schedule.

  [modelSchedule activateIn: self];

  // Finally, return our activity.

  return [self getActivity];
}

@end













