

// Paul Johnson <pauljohn@ku.edu> 2002
// This code is made available under the GPL (GNU general public license)
// with the additional proviso that any users notify the author of their
// use and share their code with the original author.  A copy of version
// 2 of the GPL is included in this directory, but you can use a newer
// GPL at your choice.  The author assumes no liability for any 
// inconvenience or damage caused by this software and offers no 
// warranty of any kind.


#import <misc.h>
#import <string.h>

#import <collections.h>
#import <simtoolsgui.h>      //pj for CREATEPROBEDISPLAY
#import <random.h>       //allows selectRandomMemberX:Y: function

#import <simtools.h>

#import "MultiGrid2d.h"	   // Multiple Occupancy Discrete2d grid
#import "MultiGridCell.h" //need protocol AppCell
#import <objc/Object.h>

// When an object is added to an empty position on the grid, a cell with a
// collection is created to hold the object. Further additions of
// objects to the cell are just added to a container.

#import "MultiGridCell.h"

@implementation MultiGrid2d


// makeOffsets, allocLattice, setSize<X,Y>, are copied from
// Grid2d.  This is a free standing class, if you want...

-makeOffsets {
  unsigned int i;

  offsets = [[self getZone] alloc: ysize * sizeof(*offsets)];

  for (i = 0; i < ysize; i++)
    offsets[i] = xsize * i;		// cache this multiplication
  return self;
}

-(id *) allocLattice {
  void * p;
  p = [[self getZone] alloc: xsize * ysize * sizeof(id)];
  memset(p, 0, xsize * ysize * sizeof (id));
  return p;
}

-setSizeX: (unsigned)x Y: (unsigned)y {
  if (lattice)
    [InvalidArgument raiseEvent: 
     "MultiGrid2d: You cannot reset the grid size after creation.  Do this before createEnd is called\n"];
  xsize = x;
  ysize = y;
  return self;
}

-(void) setCellClass: (Class)aClass {
  if (cellClass)
    [InvalidArgument raiseEvent: 
      "MultiGrid2d setCellClass: cannot set cellClass twice. Do this method before createEnd is called.\n"];
  cellClass = aClass;
}

-createEnd { 
  unsigned int x,y;
  if (xsize <= 0 || ysize <= 0)
    [InvalidCombination raiseEvent: 
			  "MultiGrid2d: Invalid size in creation: %d %d\n", xsize, ysize];
  lattice = [self allocLattice];
  [self makeOffsets];
  
  recyclerList = [List create: [self getZone]];
  
  //if no cellClass is set by now, just use the default "list" holder class
  //MultiGridCell. Have to create an instance to get the class object.
  if (!cellClass){
    [self setCellClass: [MultiGridCell class]];
  }


  //initialize with nil
  for (x=0;x< xsize; x++)
    for (y=0; y<ysize; y++)
      {
	*discrete2dSiteAt(lattice, offsets, x, y) = nil;
      }
  fprintf(stderr,"\n\n Set cell type MultiGridCell \n\n");

#ifdef KEEP_GLOBAL_LIST  
  globalList = [[[AVLSet createBegin: [self getZone]] setCompareIDs] createEnd];
#endif
  return self;
}

// Creation and parameter setting:

+create: (id) aZone setSizeX: (unsigned)x Y: (unsigned)y {
  MultiGrid2d * grid;


  fprintf(stderr,"create mg space");
  grid = [MultiGrid2d createBegin: aZone];
  [grid setSizeX: (unsigned) x Y: (unsigned) y];
  grid = [grid createEnd];
  return grid;
}


+create: (id)aZone setSizeX: (unsigned)x Y: (unsigned)y Cells: (Class)aClass {
  MultiGrid2d * grid;
  grid = [MultiGrid2d createBegin: aZone];
  [grid setSizeX: x Y: y];
  [grid setCellClass: aClass];
  grid = [grid createEnd];

  return grid;
}



-(unsigned) getSizeX {
  return xsize;
}

-(unsigned) getSizeY {
  return ysize;
}

-(id *) getLattice {
  return lattice;
}

-(long *) getOffsets 
{  
  return offsets;
}

-add: (id)anObject atX: (unsigned)x Y: (unsigned)y
{
  id cell;
  cell = *discrete2dSiteAt(lattice, offsets, x, y);

  if (!cell) 
    {						// site is empty now
      cell= [self createCellX: x Y: y];
    }
 
  [cell insertItem: anObject];		// add object to list
#ifdef KEEP_GLOBAL_LIST
  [globalList add: anObject];
#endif
  //printf("Added one to cell , now total at %d,%d is %d",x,y,[cell getCount]);

  return self; 
}


- putObject: anObject atX: (unsigned)x Y: (unsigned)y 
{
  [self add: anObject atX: x Y: y];
  return self;
}

- getCellAtX: (unsigned)x Y: (unsigned)y
{
  id cell;
  cell = *discrete2dSiteAt(lattice, offsets, x, y);
  return cell;
}

- getObjectAtX:(unsigned)x Y: (unsigned)y
{
  return [self getCellAtX: x Y: y];
}


- (long)getValueAtX: (unsigned)x Y: (unsigned)y{
  return [[self getCellAtX: x Y: y] getCellValue];
}

-(id) getFirstAtX: (unsigned)x Y: (unsigned)y 
{
  id cell,item = nil;
  cell = *discrete2dSiteAt(lattice, offsets, x, y);

  fprintf(stderr,"Grid GetFirst selecting at %d,%d \n", x, y);

  if (!cell) return nil;	// i.e. nil; cell has no list

  item =  [cell getFirst];

  //Check if cell is empty; remove list
#ifndef NO_AGGRESSIVE_RECYCLING    
  if ([cell getCount]==0){
    [recyclerList addLast: cell];
    *discrete2dSiteAt(lattice, offsets, x, y) = nil;	       // nullify the Grid value
  }
#endif

  return item;
}




-(id) getLastAtX: (unsigned)x Y: (unsigned)y 
{
  id cell,item;
  int num;

  cell = *discrete2dSiteAt(lattice, offsets, x, y);

  if (!cell) return nil;
  num = [cell getCount];

  if (num > 1)
    {
      item= [cell getLast];
      return item;
    }

  else if (num == 1) 
    {
      item = [cell getLast];
#ifndef NO_AGGRESSIVE_RECYCLING
      [recyclerList addLast: cell];
      *discrete2dSiteAt(lattice, offsets, x, y) = nil;	// zero the cell value
#endif
      return item;
    }
  else
    {
      // If we're here, num = 0 which should not happen if Aggressive REcycling is used
#ifndef NO_AGGRESSIVE_RECYCLING
      [recyclerList addLast: cell];
      *discrete2dSiteAt(lattice, offsets, x, y) = nil;	// zero the cell value
#endif
    }
  printf("MultiGrid2d: getFirst: found empty list!\n");
  return nil;
}



-(id) removeFirstAtX: (unsigned)x Y: (unsigned)y 
{
  id cell,item = nil;
  cell = *discrete2dSiteAt(lattice, offsets, x, y);

  if (!cell) return nil;
  item =  [cell removeFirst];
#ifdef KEEP_GLOBAL_LIST  
  [globalList remove: item];
#endif  

  //Check if cell is empty; remove list
#ifndef NO_AGGRESSIVE_RECYCLING    
  if ([cell getCount]==0){
    [recyclerList addLast: cell];
    *discrete2dSiteAt(lattice, offsets, x, y) = nil;	       // nullify the Grid value
  }
#endif

  return item;
}


-(id) removeLastAtX: (unsigned)x Y: (unsigned)y 
{
  id cell,item;
  int num;

  cell = *discrete2dSiteAt(lattice, offsets, x, y);

  if (!cell) return nil;	// i.e. nil; cell has no list

  num = [cell getCount];

  if (num > 1)
    {
      item= [cell removeLast ];
#ifdef KEEP_GLOBAL_LIST  
      [globalList remove: item];
#endif
      return item;
    }

  else if (num == 1) 
    {
      item = [cell removeAtOffset: num-1];
#ifdef KEEP_GLOBAL_LIST  
      [globalList remove: item];
#endif
#ifndef NO_AGGRESSIVE_RECYCLING
      [recyclerList addLast: cell];
      *discrete2dSiteAt(lattice, offsets, x, y) = nil;	// zero the cell value
#endif
      return item;
    }
  else
    {
      // If we're here, num = 0 which should not happen if Aggressive REcycling is used
#ifndef NO_AGGRESSIVE_RECYCLING
      [recyclerList addLast: cell];
      *discrete2dSiteAt(lattice, offsets, x, y) = nil;	// zero the cell value
#endif
    }
  printf("MultiGrid2d: deleteFirst: found empty list!\n");
  return nil;
}


//Note: this does not delete or remove the randomly chosen member.
//it simply gives back a pointer to it.

-(id) selectRandomMemberX: (unsigned)x Y: (unsigned)y
{
  id cell, item = nil;
  cell = *discrete2dSiteAt(lattice, offsets, x, y);

  //fprintf(stderr,"Grid selecting at %d,%d \n", x, y);
  if (!cell) return nil; 
  else
    {
      item = [cell getRandomItem];
    }  
  return item;
}

-(id) deleteObject: (id)anObject atX: (unsigned)x Y: (unsigned)y {
  id cell;
  id item=nil;

  cell = *discrete2dSiteAt(lattice, offsets, x, y);
  if (!cell) return nil;	// no list there, so return nil

  item = [cell remove: anObject];
#ifdef KEEP_GLOBAL_LIST  
  [globalList remove: item];
#endif
#ifndef NO_AGGRESSIVE_RECYCLING
  if ([cell getCount]==0) 
    {
      [recyclerList addLast: cell];
      *discrete2dSiteAt(lattice, offsets, x, y) = nil;	// zero the cell value
    }
#endif
  return item;
}


- (void) removeAllAtX: (unsigned)x Y: (unsigned)y
{
  id cell;
  cell = *discrete2dSiteAt(lattice, offsets, x, y);
  if (cell) {
    [cell removeAllItems];
#ifndef NO_AGGRESSIVE_RECYCLING   
    [recyclerList addLast: cell]; 
    *discrete2dSiteAt(lattice, offsets, x, y) = nil;
#endif
  }
}


- (void) deleteAllAtX: (unsigned)x Y: (unsigned)y
{
  id cell;
  cell = *discrete2dSiteAt(lattice, offsets, x, y);
  if (cell) {
    [cell deleteAllItems];
#ifndef NO_AGGRESSIVE_RECYCLING   
    [recyclerList addLast: cell]; 
    *discrete2dSiteAt(lattice, offsets, x, y) = nil;
#endif
  }
}

// Retrieve object data from cells:

-(unsigned) numberOfObjectsAtX: (unsigned)x Y: (unsigned)y {
  id cell;
  
  cell = *discrete2dSiteAt(lattice, offsets, x, y);
  if (cell)
    {
      return [cell getCount];
    }
  else return 0;
}


-getCollectionAtX: (unsigned)x Y: (unsigned)y 
{
  id cell=nil;
  cell=*discrete2dSiteAt(lattice, offsets, x, y);
  if (cell)  return [cell getCollection];
  else return nil;
}


// Return a set full of references to objects.
// I don't really recommend this unless you compile
// with KEEP_GLOBAL_LIST.

-(AVLSet *) getGlobalList
{

#ifdef KEEP_GLOBAL_LIST  
  return globalList;
#else
  unsigned int x,y;
  id <AppCell> cell;
  
  id combinedList = [[[AVLSet createBegin: [self getZone]] setCompareIDs] createEnd];
  
  for (x=0;x< xsize; x++)
    for (y=0; y<ysize; y++)
      {
	if ( (cell = *discrete2dSiteAt(lattice, offsets, x, y)))
	  {
	    int length = [cell getCount];
	    int i;
	    for (i = 0; i < length; i++)
	      {
		[combinedList add: [cell getAtOffset: i]];
	      }
	  }
      }
  
  return combinedList;
#endif
}


-createCellX: (unsigned)x Y: (unsigned)y
{
  id <AppCell> cell;

  if ( ([recyclerList getCount]!=0) ) 
    {
      cell=[recyclerList removeFirst];
    }
  else
    {
      cell = [cellClass createBegin: [self getZone]];
      cell = [cell createEnd];
    }

  [cell setX: x Y: y];	
  [cell initializeAllValues];
  if (cellSelector1)
    [cell perform: cellSelector1];
  if (cellSelector2)
    [cell perform: cellSelector2 with: cellArgument];
  *discrete2dSiteAt(lattice, offsets, x, y) = cell; 
  return cell;
}

- makeProbeAtX: (unsigned)x Y: (unsigned)y
{
  id obj;
  id <AppCell> cell;
  if (x >= 0
      && x < xsize
      && y >= 0
      && y < ysize)
    {
      cell = *discrete2dSiteAt(lattice, offsets, x, y);
      if(cell)
	{
	  obj = [cell getRandomItem];
	  CREATE_PROBE_DISPLAY (obj);
       	}
      else
        GUI_BEEP ();
    }
  else
    [WarningMessage
      raiseEvent:
	"MultiGrid2d: invalid coordinates to make probe (%d,%d)\n",
      x, y];
  return self;
}



- passThroughToCellCreation: (SEL)aSelector{
  cellSelector1 = aSelector;
  return self;
}


- passThroughToCellCreation: (SEL)aSelector with: arg1 {
  cellSelector2 = aSelector;
  cellArgument = arg1;
  return self;
}

- passThroughToAllCells: (SEL)aSelector
{
  unsigned int x,y;
  id cell;

  for (x=0;x< xsize; x++)
    for (y=0; y<ysize; y++)
      {
	if ((cell = *discrete2dSiteAt(lattice, offsets, x, y)))
	  [cell perform: aSelector];
      }
  return self;
}


- passThroughToAllCells: (SEL)aSelector with: arg1
{
  unsigned int x,y;
  id <AppCell> cell;

  for (x=0;x< xsize; x++)
    for (y=0; y<ysize; y++)
      {
	if ( (cell = *discrete2dSiteAt(lattice, offsets, x, y)))
	  [cell perform: aSelector with: arg1];
      }
  return self;
}


- passThroughToCell: (SEL)aSelector atX: (unsigned)x Y: (unsigned)y
{
  id <AppCell> cell;
  if ( (cell = *discrete2dSiteAt(lattice, offsets, x, y)))
    [cell perform: aSelector];
  return self;
}


- passThroughToCell: (SEL)aSelector with: arg1 atX: (unsigned)x Y: (unsigned)y;
{
  id <AppCell> cell;
  if ( (cell = *discrete2dSiteAt(lattice, offsets, x, y)))
    [cell perform: aSelector with: arg1];
  return self;
}


// Tell all objects in the grid to do something.
- passThroughToObjects: (SEL)aSelector
{
#ifdef KEEP_GLOBAL_LIST  
  [globalList forEach: aSelector];
#else
  unsigned int x,y;
  id <AppCell> cell;
  
  for (x=0;x< xsize; x++)
    for (y=0; y<ysize; y++)
      {
	if ( (cell = *discrete2dSiteAt(lattice, offsets, x, y)))
	  [cell performForEach: aSelector];
      }
#endif
  
  return self;
}

- passThroughToObjects: (SEL)aSelector with: arg1
{
#ifdef KEEP_GLOBAL_LIST  
  [globalList forEach: aSelector: arg1];
#else
  unsigned int x,y;
  id <AppCell> cell;
  
  for (x=0;x< xsize; x++)
    for (y=0; y<ysize; y++)
      {
	if ( (cell = *discrete2dSiteAt(lattice, offsets, x, y)))
	  [cell performForEach: aSelector with: arg1];
      }
#endif
  return self;
}

// Tell all objects (x,y) to do something.
- passThroughToObjects: (SEL)aSelector atX: (unsigned)x Y: (unsigned)y
{
  id <AppCell> cell;
  
  if ( (cell = *discrete2dSiteAt(lattice, offsets, x, y)))
    [cell performForEach: aSelector];
  
  return self;
}

- passThroughToObjects: (SEL)aSelector with: arg1 atX: (unsigned)x Y: (unsigned)y
{
  id <AppCell> cell;
  
  if ( (cell = *discrete2dSiteAt(lattice, offsets, x, y)))
    [cell performForEach: aSelector with: arg1];

  return self;
}

-_I_Using { return self; }

//inserted to make this write itself to a file!

//  -writeValuesToFile
//  {
//     id gridfiler;
//     char filename[40];
//     sprintf(filename, "%05ldMultiGrid.txt", getCurrentTime());
//     gridfiler=[Int2dFiler createBegin: [self getZone]];
//     [gridfiler setDiscrete2dToFile: self];
//     [gridfiler setValueMessage: M(getOutputValue)];
//     gridfiler=[gridfiler createEnd];
    
//     [gridfiler fileTo: filename];
//     return self;
//  }


@end










