
// Paul Johnson <pauljohn@ku.edu> 2002
// This code is made available under the GPL (GNU general public license)
// with the additional proviso that any users notify the author of their
// use and share their code with the original author.  A copy of version
// 2 of the GPL is included in this directory, but you can use a newer
// GPL at your choice.  The author assumes no liability for any 
// inconvenience or damage caused by this software and offers no 
// warranty of any kind.


#import <space.h>
#import <objectbase/SwarmObject.h>
#import <collections.h>


@protocol AppCell

//if you build a sublcass from MultiGridCell,
//generally, you don't have to override these ones.
- getCollection;
- setX: (unsigned) theX Y: (unsigned) theY;
- setX: (unsigned) theX;
- setY: (unsigned) theY;
- (unsigned) getX;
- (unsigned) getY;
- createEnd; 

// Definitely you need to override these if you build a subclass,
// since they use the special features of your collection
- (id) createCollection;
- (long) getCellValue;
// Swarm insert methods return void, but I leave the right
// to return a status indicator by making it bool.
- (BOOL) insertItem:  obj ;
// Some collections need keys to insert, so make room for that
- (BOOL) insertItem: obj1 Key: obj2;

// This is vital for "recycling" of cells. It takes an old cell and
// sets its ivars back to a virgin state.
- initializeAllValues;


//If your cell class uses a Swarm collection, you probably don't
//need to override these methods from MultiGridCell, since they
//use the standard Swarm collections commands to remove,add,etc.

- (unsigned) getCount;
- getFirst;
- getLast;
- getRandomItem;
- removeRandomItem;
- removeFirst;
- removeLast;
- remove: (id)obj;
- (void) removeAllItems;
- (void) deleteAllItems;
- removeAtOffset: (unsigned)offset;
- getAtOffset: (unsigned)offset;

- performForEach: (SEL)aSel;
- performForEach: (SEL)aSel with: anObject;


//inherit these from Object, no need to override
- perform:(SEL)aSel;
- perform:(SEL)aSel with: anObject;

@end


@interface MultiGridCell: SwarmObject <AppCell>
{
   id  list;
   int x,y;
}

- createEnd;
- getCollection;
- setX: (unsigned)theX Y: (unsigned)theY;

- setX: (unsigned)theX;
- setY: (unsigned)theY;
- (unsigned)getX;
- (unsigned)getY;

- (unsigned)getCount;

- createCollection;
- (long) getCellValue; 

- (BOOL) insertItem:  obj ;
- (BOOL) insertItem: obj1 Key: obj2;

- getRandomItem;
- (id) getFirst;
- (id) getLast;

- removeRandomItem;
- remove: (id) obj;
- removeAtOffset: (unsigned)offset;
- removeFirst;
- removeLast;

- (void) removeAllItems;
- (void) deleteAllItems;

- (id) getAtOffset: (unsigned)offset;
- initializeAllValues;

- performForEach: (SEL)aSel;
- performForEach: (SEL)aSel with: anObject;

@end










