// Paul Johnson <pauljohn@ku.edu> 2002
// This code is made available under the GPL (GNU general public license)
// with the additional proviso that any users notify the author of their
// use and share their code with the original author.  A copy of version
// 2 of the GPL is included in this directory, but you can use a newer
// GPL at your choice.  The author assumes no liability for any 
// inconvenience or damage caused by this software and offers no 
// warranty of any kind.


#import "MultiGridCell.h"
#import <math.h>
#import <misc.h>
#import <random.h>  //uniformUnsRand

@implementation MultiGridCell

- createEnd
{
  [self createCollection];
  return [super createEnd];
}

- setX: (unsigned)theX Y: (unsigned)theY
{
  x=theX;
  y=theY;
  return self;
}

- setX: (unsigned)theX
{
    x = theX;
    return self;
}

- setY: (unsigned)theY
{
    y = theY;
    return self;
}


-(id) getCollection
{
    return list;
}

- (unsigned) getX
{ 
    return x;
}

- (unsigned) getY
{
    return y;
}

-(unsigned) getCount
{
  return [list getCount];
}


- (id) createCollection 
{
  //raiseEvent (WarningMessage,"The MultiGrid2d is using Swarm Lists to hold objects.  If you want a fancy collection, customize it!  See docs in MultiGrid2d");
  list = [List create: [self getZone]];
  return list;
}


//Just returns 0. If you need something that has substantive
// meaning, override it in your subclass.
-(long) getCellValue{
  return (long) 0;
}

- (BOOL) insertItem: obj 
{
  [list addFirst: obj];
  return YES;
} 


- (BOOL) insertItem: obj1 Key: obj2
{
  raiseEvent(InvalidArgument,  "MultiGridCell: lists don't use keys");
  return NO;
} 

- getRandomItem 
{
  unsigned num = [list getCount];
  id item = nil;
  num = [list getCount];
  // fprintf(stderr,"MultiGridCell %d,%d getRandom CellSize: %d \n", x,y, num);
  if (num == 0) return nil;
  else if (num == 1) return [list atOffset:0];
  else
    item = [list atOffset: [uniformUnsRand getUnsignedWithMin: 0 withMax: num-1] ];
  return item;
}

- getFirst
{
  return [list getFirst];
}
- getLast
{
  return [list getLast];
}


- removeRandomItem 
{
  unsigned int i;
  id obj = nil;
  unsigned num = [list getCount];
   if (num == 0) return nil;
  else if (num == 1) return [list atOffset:0];
  else if (num > 1)
    {
      id obj;
      id index = [list begin: [self getZone]];
      unsigned xx = [uniformUnsRand getUnsignedWithMin: 0 withMax: [list getCount]-1];
      for ( i = 0; i < xx; i++){
	[index next];
      }
      obj = [index remove];
      [index drop];
    } 
   return obj;
}

- remove: anObj
{
  id obj = [list remove: anObj];
  return obj;
}

- (void) removeAllItems 
{
  [list removeAll];
}

- (id) removeFirst
{
  return [list removeFirst];
}

- (id) removeLast
{
  return [list removeLast];
}

- (void) deleteAllItems 
{
  [list deleteAll];
}

//remove the Nth item, N=0,...,count-1
- removeAtOffset: (unsigned)offset 
{
  id index = [list begin: [self getZone]];
  id obj;
  for ( offset++; offset > 0; offset--)  {
    [index next];
  } 
  obj = [index remove];
  [index drop];
  return obj;
}

- getAtOffset: (unsigned)offset
{
  id obj  = [list atOffset: offset]; 
  return obj;
}

- performForEach: (SEL)aSel
{
  [list forEach: aSel];
  return self;
}

- performForEach: (SEL) aSel with: anObject
{
  [list forEach: aSel:anObject];
  return self;
}


// Look Here! This method does nothing, because by default the cells
// don't have any values that need initializing.  If you cells have
// values in them, be sure to override this!

- initializeAllValues{
  return nil;
}

@end 











