//MyArguments.m

// Paul Johnson <pauljohn@ku.edu> 2002
// This code is made available under the GPL (GNU general public license)
// with the additional proviso that any users notify the author of their
// use and share their code with the original author.  A copy of version
// 2 of the GPL is included in this directory, but you can use a newer
// GPL at your choice.  The author assumes no liability for any 
// inconvenience or damage caused by this software and offers no 
// warranty of any kind.


#import "MyArguments.h"
#import <stdlib.h>
#import <objectbase.h>  //for VarProbe usage in functions
#import <simtools.h>  // need swarmGUIMode below
#import <misc.h> //because strtod is used

id
makeProbe (id obj, const char *ivarName)
{
  id probe = [VarProbe createBegin: [obj getZone]];

  [probe setProbedClass: [obj getClass]];
  [probe setProbedVariable: ivarName];
  return [probe createEnd];
}

double
getDouble (id obj, const char *ivarName)
{
  id probe = makeProbe (obj, ivarName);
  double ret = [probe probeAsDouble: obj];
  [probe drop];
  return ret;
}

int
getInt (id obj, const char *ivarName)
{
  id probe = makeProbe (obj, ivarName);
  int ret = [probe probeAsInt: obj];
  [probe drop];
  return ret;
}

@implementation MyArguments

+ createBegin: aZone
 {
   static struct argp_option options[] = { 

     {"numPTwo",'p',"P", 0, "Set numPTwo", 5},
     {"numPPL",'n',"N", 0, "Set numPPL", 6},
     {"inputfile", 'I',"filename",0,"set fn",7},
     {"run",'R',"RunNumber",0,"Run is...",8},
     {"vision",'v',"VisionRadius",0,"Vision to radius squares",11},
     {"citizensMove",'m',"0 or 1", 0, "Set citizensMove",12},
     {"worldsize",       'w',"WorldSize",0,"Dimension w and h",13},
     {"wrap", 'r', "0 or 1", 0, "wrapping around neighborhoods", 14},
     {"SYNC", 'Z', "0 or 1", 0, "SYNCHRONOUS UPDATING",15},
     {"RAND", 'a', "0 or 1", 0, "RANDOMIZED Agent List",16},
     {"probYES", 'Y', "probability", 0, "probability of YES", 17},
     {"duration", 'd', "N of timesteps", 0, "length of simulation",18},
     {"UNATTENDED", 'U', "0 or 1", 0, "Run GUI without user interaction",19},
     {"GUIshots", 'G', "0 or 1", 0, "Save images in PNG format",20},
     { 0 }
   };
   
   MyArguments *obj = [super createBegin: aZone];

   [obj addOptions: options];

   obj->worldXSize = 40;
   obj->worldYSize = 40;
   obj->visionRadius = 77;
   obj->citizensMove = 0;
   obj->randomSeed = 4;
   obj->numPPL = 1600;
   obj->numPTwo = 0;
   obj->wrapAround = NO;
   obj->SYNCHRONOUS = YES;
   obj->RANDOMIZED = YES;
   obj->GUIshots = NO;
   obj->UNATTENDED = NO;
   obj->probYES=0.5;
   obj->run = -1;
   obj->duration = 30;
   return obj;
 }

- (int)parseKey: (int)key arg: (const char*)arg
 { 
   if (key == 'm')
     {
       citizensMove = atoi(arg);
       return 0;
     }      
   else if (key == 'n')
     {
       numPPL = atoi(arg);
       return 0;
     }
   else if (key == 'v')
     {
       visionRadius = atoi(arg);
       return 0;
     } 
    else if (key == 'w')
    {
      worldXSize= atoi(arg);
      worldYSize= worldXSize;
      return 0;
    }

  //  else if (key == 'S')
//     {
//       seedArg = arg;
//       randomSeed=atoi(seedArg);
//       return 0;
//     }
   else if (key == 'd')
     {
       duration = atoi(arg);
       return 0;
     }


   else if (key == 'I')
     {
       printf("here is an I \n");
       return 0;
     }
   else if (key == 'R')
     {
       run = atoi(arg);
       return 0;
     }
   else if (key == 'r')
     {
       wrapAround = atoi(arg);
       return 0;
     }
   else if (key == 'Z')
     {
       SYNCHRONOUS = atoi(arg);
       return 0;
     }
   else if (key == 'a')
     {
       RANDOMIZED = atoi(arg);
       return 0;
     }

   else if (key == 'Y')
     {
       probYES = strtod (arg, 0);
       return 0;
     }
   else if (key == 'G')
     {
       GUIshots = atoi(arg);
       return 0;
     }
   else if (key == 'U')
     {
       UNATTENDED = atoi(arg);
       return 0;
     }

   else
     return [super parseKey: key arg: arg];
 }


- init
{
  id probeMap;

  // [self createCharDate];

  //  charDate = [TimeStamp getTimeStamp];
  //  printf("TimeStamp says current time is is %s \n",charDate);
  randomSeed = [randomGenerator getInitialSeed];
  
  if (swarmGUIMode == 1)
    {
      probeMap = [EmptyProbeMap createBegin: [self getZone]];
      [probeMap setProbedClass: [self class]];
      probeMap = [probeMap createEnd];
  
      [probeMap addProbe: [probeLibrary getProbeForVariable: "citizensMove"
					inClass: [self class]]];
      [probeMap addProbe: [probeLibrary getProbeForVariable: "visionRadius"
					inClass: [self class]]];
      [probeMap addProbe: [probeLibrary getProbeForVariable: "worldXSize"
					inClass: [self class]]];
      [probeMap addProbe: [probeLibrary getProbeForVariable: "worldYSize"
					inClass: [self class]]];
      [probeMap addProbe: [probeLibrary getProbeForVariable: "numPPL"
					inClass: [self class]]]; 
      [probeMap addProbe: [probeLibrary getProbeForVariable: "probYES"
				     inClass: [self class]]]; 
      [probeMap addProbe: [probeLibrary getProbeForVariable: "randomSeed"
					inClass: [self class]]];
      [probeMap addProbe: [probeLibrary getProbeForVariable: "wrapAround"
					inClass: [self class]]];
      [probeMap addProbe: [probeLibrary getProbeForVariable: "SYNCHRONOUS"
					inClass: [self class]]]; 
      [probeMap addProbe: [probeLibrary getProbeForVariable: "RANDOMIZED"
					inClass: [self class]]];

      [probeMap addProbe: [probeLibrary getProbeForVariable: "GUIshots"
				     inClass: [self class]]];
      [probeLibrary setProbeMap: probeMap For: [self class]];

      

    }
  return self;
}


- (BOOL)getWrapAround
{
  return wrapAround;
}


- (BOOL)getSYNCHRONOUS
{
  return SYNCHRONOUS;
}

- (BOOL)getRANDOMIZED
{
  return RANDOMIZED;
}
- (int)getVisionRadiusArg
{
  return visionRadius;
}

- (int)getRunArg
{
  return run;
}


- sayHello
{
    printf("MyArguments says hello to you");
    return self;
}

@end
































